/**
 * Define outbound to load entity refs
 * @function refOutbound
 * @param {ClayResource} refResource - Resource of entities
 * @returns {function} Outbound function
 */
'use strict'

const {isEntity} = require('clay-entity')
const {parse: parseRef, refTo} = require('clay-resource-ref')
const {LogPrefixes} = require('clay-constants')
const clayResourceName = require('clay-resource-name')
const {RESOURCE_PREFIX} = LogPrefixes

/** @lends refOutbound */
function refOutbound (refResource) {
  const refResourceName = String(clayResourceName(refResource))

  async function outboundAttribute (resourceName, name, attribute, actionContext) {
    const {
      skipResolvingRefFor = [],
      ignoreCached = false,
      resolvedRefs = {},
      suppressWarning = false,
    } = actionContext
    const skip = skipResolvingRefFor.includes(refResourceName) || skipResolvingRefFor.includes(name)
    if (skip) {
      return attribute
    }

    if (Array.isArray(attribute)) {
      return Promise.all(
        attribute.map((attribute) => {
          if (Array.isArray(attribute)) {
            // Skip for nested array
            return attribute
          }
          return outboundAttribute(resourceName, name, attribute, actionContext)
        })
      )
    }
    if (isEntity(attribute)) {
      if (!attribute.$$as) {
        if (!suppressWarning) {
          console.warn(`${RESOURCE_PREFIX} You cannot use "${attribute}" as ref of "${name}" since it is not annotated`)
        }
        return attribute
      }
      const hit = String(attribute.$$as) === refResourceName
      if (hit) {
        attribute = {$ref: refTo(refResource, attribute.id)}
      }
    }
    const isNullRef = attribute && attribute.hasOwnProperty('$ref') && attribute.$ref === null
    if (isNullRef) {
      return null
    }
    const $ref = parseRef(attribute && attribute.$ref)
    if (!$ref) {
      return attribute
    }
    const hit = String($ref.resource) === refResourceName
    if (hit) {
      const refKey = refTo(refResource, $ref.id)
      const resolved = resolvedRefs[refKey]
      if (resolved) {
        return resolved
      }
      const found = refResource.one($ref.id, {
        ignoreCached,
        skipResolvingRefFor: [...skipResolvingRefFor, resourceName].filter(Boolean),
        fromAction: actionContext.action
      })
      resolvedRefs[refKey] = found
      return found
    }

    return attribute
  }

  /**
   * @function refOutbound
   * @param {Entity[]} - Entities
   * @param {Object} [actionContext={}]
   * @returns {Promise.<Entity[]>}
   */
  return async function outbound (resource, entities, actionContext = {}) {
    const resourceName = String(clayResourceName(resource))
    for (const entity of entities) {
      const names = Object.keys(entity).filter((name) => !/^_/.test(name))
      for (const name of names) {
        const attribute = entity[name]
        entity[name] = await outboundAttribute(resourceName, name, attribute, actionContext)
      }
    }
    return entities
  }
}

module.exports = refOutbound
